/****************************************************************************
**  Copyright (c) 2011-2012                                                **
**  German Research School for Simulation Sciences GmbH					   **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef LWM2_MPI_ADAPTER_H
#define LWM2_MPI_ADAPTER_H

#include <stdio.h>

#include "../util/util_time_measurement.h"

#include "mpi/lwm2_mpireg.h"

#include "../lwm2-signal.h"
#include "../util/lwm2-log.h"
#include "../lwm2-storage.h"

typedef struct
{
   unsigned short type;

   unsigned short group;

	/**
	 * Number of bytes sent by p2p functions
	 */
	size_t bytes_sent;
	size_t bytes_recv;
} region_information_t;

/**
 * Initialize mpi adapter
 */ 

extern int lwm2_mpi_rank;

extern timestamp_t lwm2_mpi_start_time;
extern timestamp_t lwm2_mpi_end_time;
extern timestamp_t lwm2_mpi_total_time;

extern int lwm2_mpi_comm_size ;

//extern sig_atomic_t lwm2_cuda_active_function;

int lwm2_mpi_init();

void lwm2_mpi_finalize();

void lwm2_proc_init();

#define HOSTNAME_LENGTH 64


/*****************************************/


/* Logic to detect whether MPI calling frequency is too high */
extern const timespan_t lwm2_mpi_time_to_check; /* second */

/* Local counts */
extern timestamp_t lwm2_mpi_last_function;
extern unsigned int lwm2_mpi_last_function_count[LWM2_MPI_GROUPCOUNT];

/* Global counts */
extern unsigned int lwm2_mpi_global_function_count;

/* Treshold to stay below */
extern unsigned int lwm2_mpi_function_treshold;

extern int lwm2_app_ended;
extern int lwm2_mpi_app;

extern timestamp_t lwm2_total_merge_time;

#ifdef MPI_DIRECT_SAMPLING
extern timestamp_t lwm2_mpi_total_time;
extern timestamp_t lwm2_reduced_total_lwm2_time;
extern timestamp_t lwm2_reduced_total_mpi_time;
#endif


/*****************************************/

/**
 * 
 */

void lwm2_mpi_sum_mpi_time();

double lwm2_get_time_till_time_slice();

static inline region_information_t lwm2_mpi_enter(const unsigned short function_idx, const unsigned short group)
{
   region_information_t info;     

   info.type = function_idx;
   info.group = group;
   info.bytes_recv = info.bytes_sent = 0;
   
   if(lwm2_app_ended == 1)
   {
	   return info;
   }

   lwm2_storage_set_flag_mpi();
   
   if(group == LWM2_MPI_GROUP_IO)
   {
	   lwm2_storage_set_flag_mpi_io();
   }
   
   if(group == LWM2_MPI_GROUP_P2P)
   {
	   lwm2_storage_set_flag_mpi_p2p();
   }
   
   if(group == LWM2_MPI_GROUP_COLL)
   {
	   lwm2_storage_set_flag_mpi_coll();
   }
   
#ifdef MPI_DIRECT_SAMPLING
   lwm2_mpi_start_time = util_get_timestamp();
#endif
   
#ifdef TREAT_CUDA_AS_MPI
   lwm2_storage_set_flag_cuda();
#endif

   LWM2LOG(5, "[mpi]\tRegion enter for region type %d\n", function_idx);

   return info;
}

/**
 * 
 */
static inline void lwm2_mpi_leave(region_information_t* info)
{
   if(lwm2_app_ended == 1)
   {
   		return;
   }

	if(lwm2_sys_service_active != 0)
	{
		return;
	}
#ifdef TREAT_CUDA_AS_MPI
	lwm2_storage_reset_flag_cuda();
#endif
	
#ifdef MPI_DIRECT_SAMPLING
	lwm2_mpi_end_time = util_get_timestamp();
	lwm2_mpi_total_time += lwm2_mpi_end_time - lwm2_mpi_start_time;
#endif

   if(info->group == LWM2_MPI_GROUP_P2P)
   {
	   if(info->bytes_sent != 0)
	   {
		   lwm2_storage_add_p2p_sent_bytes(info->bytes_sent);
		   lwm2_storage_increment_p2p_sent_msgs();
	   }
	   
	   if(info->bytes_recv != 0)
	   {
		   lwm2_storage_add_p2p_recv_bytes(info->bytes_recv);
		   lwm2_storage_increment_p2p_recv_msgs();
	   }
	   
	   lwm2_storage_reset_flag_mpi_p2p();
   }
   
   if(info->group == LWM2_MPI_GROUP_COLL)
   {
	   lwm2_storage_add_coll_sent_bytes(info->bytes_sent);
	   lwm2_storage_add_coll_recv_bytes(info->bytes_recv);
	   lwm2_storage_increment_coll_calls();
	   
	   lwm2_storage_reset_flag_mpi_coll();
   }
   
   if(info->group == LWM2_MPI_GROUP_IO)
   {
	   lwm2_storage_reset_flag_mpi_io();
	   
	   if(info->bytes_sent != 0)
	   {
		   lwm2_storage_add_io_write_bytes(info->bytes_sent);
	   }
	   
	   if(info->bytes_recv != 0)
	   {
		   lwm2_storage_add_io_read_bytes(info->bytes_recv);
	   }
   }
	 
   lwm2_storage_reset_flag_mpi();
   LWM2LOG(5, "[mpi]\tRegion exit for region type %d\n", info->type);
}

void lwm2_start_pulse_thread();


#endif
