/****************************************************************************
**  Wrapper originally built for:										
**  **
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  LWM2
**  **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2010                                                     **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/**
 * @file epk_mpireg.c
 *
 * @brief Registration of MPI regions
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "lwm2_mpireg.h"

/**
 * MPI operation event type
 */
typedef struct
{
  char*        name;
  unsigned int type;
} EpkMPItype;

/** MPI point-to-point function types
 * @note contents must be alphabetically sorted */
static const EpkMPItype epk_mpi_pt2pt[] = {
  { "MPI_Bsend",            LWM2_MPI_TYPE__SEND                               },
  { "MPI_Ibsend",           LWM2_MPI_TYPE__SEND                               },
  { "MPI_Irsend",           LWM2_MPI_TYPE__SEND                               },
  { "MPI_Isend",            LWM2_MPI_TYPE__SEND                               },
  { "MPI_Issend",           LWM2_MPI_TYPE__SEND                               },
  { "MPI_Recv",             LWM2_MPI_TYPE__RECV                               },
  { "MPI_Rsend",            LWM2_MPI_TYPE__SEND                               },
  { "MPI_Send",             LWM2_MPI_TYPE__SEND                               },
  { "MPI_Sendrecv",         LWM2_MPI_TYPE__SEND + LWM2_MPI_TYPE__RECV         },
  { "MPI_Sendrecv_replace", LWM2_MPI_TYPE__SEND + LWM2_MPI_TYPE__RECV         },
  { "MPI_Ssend",            LWM2_MPI_TYPE__SEND                               },
  { "MPI_Start",            LWM2_MPI_TYPE__SEND                               },
  { "MPI_Startall",         LWM2_MPI_TYPE__SEND                               },
  { "MPI_Test",             LWM2_MPI_TYPE__RECV                               },
  { "MPI_Testall",          LWM2_MPI_TYPE__RECV                               },
  { "MPI_Testany",          LWM2_MPI_TYPE__RECV                               },
  { "MPI_Testsome",         LWM2_MPI_TYPE__RECV                               },
  { "MPI_Wait",             LWM2_MPI_TYPE__RECV                               },
  { "MPI_Waitall",          LWM2_MPI_TYPE__RECV                               },
  { "MPI_Waitany",          LWM2_MPI_TYPE__RECV                               },
  { "MPI_Waitsome",         LWM2_MPI_TYPE__RECV                               },
};

/** MPI collective function types
 * @note contents must be alphabetically sorted */
static const EpkMPItype epk_mpi_colls[] = {
  { "MPI_Allgather",            LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Allgatherv",           LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Allreduce",            LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Alltoall",             LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Alltoallv",            LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Alltoallw",            LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Barrier",              LWM2_COLL_TYPE__BARRIER                 },
  { "MPI_Bcast",                LWM2_COLL_TYPE__ONE2ALL                 },
  { "MPI_Cart_create",          LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Cart_sub",             LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Comm_create",          LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Comm_dup",             LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Comm_free",            LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Comm_split",           LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Exscan",               LWM2_COLL_TYPE__PARTIAL                 },
  { "MPI_Finalize",             LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Gather",               LWM2_COLL_TYPE__ALL2ONE                 },
  { "MPI_Gatherv",              LWM2_COLL_TYPE__ALL2ONE                 },
  { "MPI_Init",                 LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Init_thread",          LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Reduce",               LWM2_COLL_TYPE__ALL2ONE                 },
  { "MPI_Reduce_scatter",       LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Reduce_scatter_block", LWM2_COLL_TYPE__ALL2ALL                 },
  { "MPI_Scan",                 LWM2_COLL_TYPE__PARTIAL                 },
  { "MPI_Scatter",              LWM2_COLL_TYPE__ONE2ALL                 },
  { "MPI_Scatterv",             LWM2_COLL_TYPE__ONE2ALL                 },
  { "MPI_Win_create",           LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Win_fence",            LWM2_COLL_TYPE__IMPLIED                 },
  { "MPI_Win_free",             LWM2_COLL_TYPE__IMPLIED                 }
};

/* Regions depicting currently active MPI function */
int lwm2_mpi_region[LWM2_MPI_FUNCCOUNT];

/* Names for all MPI regions */
char* lwm2_mpi_region_names[LWM2_MPI_FUNCCOUNT + 1] = {
  "MPI_Abort",
  "MPI_Accumulate",
  "MPI_Add_error_class",
  "MPI_Add_error_code",
  "MPI_Add_error_string",
  "MPI_Address",
  "MPI_Allgather",
  "MPI_Allgatherv",
  "MPI_Alloc_mem",
  "MPI_Allreduce",
  "MPI_Alltoall",
  "MPI_Alltoallv",
  "MPI_Alltoallw",
  "MPI_Attr_delete",
  "MPI_Attr_get",
  "MPI_Attr_put",
  "MPI_Barrier",
  "MPI_Bcast",
  "MPI_Bsend",
  "MPI_Bsend_init",
  "MPI_Buffer_attach",
  "MPI_Buffer_detach",
  "MPI_Cancel",
  "MPI_Cart_coords",
  "MPI_Cart_create",
  "MPI_Cart_get",
  "MPI_Cart_map",
  "MPI_Cart_rank",
  "MPI_Cart_shift",
  "MPI_Cart_sub",
  "MPI_Cartdim_get",
  "MPI_Close_port",
  "MPI_Comm_accept",
  "MPI_Comm_c2f",
  "MPI_Comm_call_errhandler",
  "MPI_Comm_compare",
  "MPI_Comm_connect",
  "MPI_Comm_create",
  "MPI_Comm_create_errhandler",
  "MPI_Comm_create_keyval",
  "MPI_Comm_delete_attr",
  "MPI_Comm_disconnect",
  "MPI_Comm_dup",
  "MPI_Comm_f2c",
  "MPI_Comm_free",
  "MPI_Comm_free_keyval",
  "MPI_Comm_get_attr",
  "MPI_Comm_get_errhandler",
  "MPI_Comm_get_name",
  "MPI_Comm_get_parent",
  "MPI_Comm_group",
  "MPI_Comm_join",
  "MPI_Comm_rank",
  "MPI_Comm_remote_group",
  "MPI_Comm_remote_size",
  "MPI_Comm_set_attr",
  "MPI_Comm_set_errhandler",
  "MPI_Comm_set_name",
  "MPI_Comm_size",
  "MPI_Comm_spawn",
  "MPI_Comm_spawn_multiple",
  "MPI_Comm_split",
  "MPI_Comm_test_inter",
  "MPI_Dims_create",
  "MPI_Dist_graph_create",
  "MPI_Dist_graph_create_adjacent",
  "MPI_Dist_graph_neighbors",
  "MPI_Dist_graph_neighbors_count",
  "MPI_Errhandler_create",
  "MPI_Errhandler_free",
  "MPI_Errhandler_get",
  "MPI_Errhandler_set",
  "MPI_Error_class",
  "MPI_Error_string",
  "MPI_Exscan",
  "MPI_File_c2f",
  "MPI_File_call_errhandler",
  "MPI_File_close",
  "MPI_File_create_errhandler",
  "MPI_File_delete",
  "MPI_File_f2c",
  "MPI_File_get_amode",
  "MPI_File_get_atomicity",
  "MPI_File_get_byte_offset",
  "MPI_File_get_errhandler",
  "MPI_File_get_group",
  "MPI_File_get_info",
  "MPI_File_get_position",
  "MPI_File_get_position_shared",
  "MPI_File_get_size",
  "MPI_File_get_type_extent",
  "MPI_File_get_view",
  "MPI_File_iread",
  "MPI_File_iread_at",
  "MPI_File_iread_shared",
  "MPI_File_iwrite",
  "MPI_File_iwrite_at",
  "MPI_File_iwrite_shared",
  "MPI_File_open",
  "MPI_File_preallocate",
  "MPI_File_read",
  "MPI_File_read_all",
  "MPI_File_read_all_begin",
  "MPI_File_read_all_end",
  "MPI_File_read_at",
  "MPI_File_read_at_all",
  "MPI_File_read_at_all_begin",
  "MPI_File_read_at_all_end",
  "MPI_File_read_ordered",
  "MPI_File_read_ordered_begin",
  "MPI_File_read_ordered_end",
  "MPI_File_read_shared",
  "MPI_File_seek",
  "MPI_File_seek_shared",
  "MPI_File_set_atomicity",
  "MPI_File_set_errhandler",
  "MPI_File_set_info",
  "MPI_File_set_size",
  "MPI_File_set_view",
  "MPI_File_sync",
  "MPI_File_write",
  "MPI_File_write_all",
  "MPI_File_write_all_begin",
  "MPI_File_write_all_end",
  "MPI_File_write_at",
  "MPI_File_write_at_all",
  "MPI_File_write_at_all_begin",
  "MPI_File_write_at_all_end",
  "MPI_File_write_ordered",
  "MPI_File_write_ordered_begin",
  "MPI_File_write_ordered_end",
  "MPI_File_write_shared",
  "MPI_Finalize",
  "MPI_Finalized",
  "MPI_Free_mem",
  "MPI_Gather",
  "MPI_Gatherv",
  "MPI_Get",
  "MPI_Get_address",
  "MPI_Get_count",
  "MPI_Get_elements",
  "MPI_Get_processor_name",
  "MPI_Get_version",
  "MPI_Graph_create",
  "MPI_Graph_get",
  "MPI_Graph_map",
  "MPI_Graph_neighbors",
  "MPI_Graph_neighbors_count",
  "MPI_Graphdims_get",
  "MPI_Grequest_complete",
  "MPI_Grequest_start",
  "MPI_Group_c2f",
  "MPI_Group_compare",
  "MPI_Group_difference",
  "MPI_Group_excl",
  "MPI_Group_f2c",
  "MPI_Group_free",
  "MPI_Group_incl",
  "MPI_Group_intersection",
  "MPI_Group_range_excl",
  "MPI_Group_range_incl",
  "MPI_Group_rank",
  "MPI_Group_size",
  "MPI_Group_translate_ranks",
  "MPI_Group_union",
  "MPI_Ibsend",
  "MPI_Info_c2f",
  "MPI_Info_create",
  "MPI_Info_delete",
  "MPI_Info_dup",
  "MPI_Info_f2c",
  "MPI_Info_free",
  "MPI_Info_get",
  "MPI_Info_get_nkeys",
  "MPI_Info_get_nthkey",
  "MPI_Info_get_valuelen",
  "MPI_Info_set",
  "MPI_Init",
  "MPI_Init_thread",
  "MPI_Initialized",
  "MPI_Intercomm_create",
  "MPI_Intercomm_merge",
  "MPI_Iprobe",
  "MPI_Irecv",
  "MPI_Irsend",
  "MPI_Is_thread_main",
  "MPI_Isend",
  "MPI_Issend",
  "MPI_Keyval_create",
  "MPI_Keyval_free",
  "MPI_Lookup_name",
  "MPI_Op_c2f",
  "MPI_Op_commutative",
  "MPI_Op_create",
  "MPI_Op_f2c",
  "MPI_Op_free",
  "MPI_Open_port",
  "MPI_Pack",
  "MPI_Pack_external",
  "MPI_Pack_external_size",
  "MPI_Pack_size",
  "MPI_Pcontrol",
  "MPI_Probe",
  "MPI_Publish_name",
  "MPI_Put",
  "MPI_Query_thread",
  "MPI_Recv",
  "MPI_Recv_init",
  "MPI_Reduce",
  "MPI_Reduce_local",
  "MPI_Reduce_scatter",
  "MPI_Reduce_scatter_block",
  "MPI_Register_datarep",
  "MPI_Request_c2f",
  "MPI_Request_f2c",
  "MPI_Request_free",
  "MPI_Request_get_status",
  "MPI_Rsend",
  "MPI_Rsend_init",
  "MPI_Scan",
  "MPI_Scatter",
  "MPI_Scatterv",
  "MPI_Send",
  "MPI_Send_init",
  "MPI_Sendrecv",
  "MPI_Sendrecv_replace",
  "MPI_Sizeof",
  "MPI_Ssend",
  "MPI_Ssend_init",
  "MPI_Start",
  "MPI_Startall",
  "MPI_Status_c2f",
  "MPI_Status_f2c",
  "MPI_Status_set_cancelled",
  "MPI_Status_set_elements",
  "MPI_Test",
  "MPI_Test_cancelled",
  "MPI_Testall",
  "MPI_Testany",
  "MPI_Testsome",
  "MPI_Topo_test",
  "MPI_Type_c2f",
  "MPI_Type_commit",
  "MPI_Type_contiguous",
  "MPI_Type_create_darray",
  "MPI_Type_create_f90_complex",
  "MPI_Type_create_f90_integer",
  "MPI_Type_create_f90_real",
  "MPI_Type_create_hindexed",
  "MPI_Type_create_hvector",
  "MPI_Type_create_indexed_block",
  "MPI_Type_create_keyval",
  "MPI_Type_create_resized",
  "MPI_Type_create_struct",
  "MPI_Type_create_subarray",
  "MPI_Type_delete_attr",
  "MPI_Type_dup",
  "MPI_Type_extent",
  "MPI_Type_f2c",
  "MPI_Type_free",
  "MPI_Type_free_keyval",
  "MPI_Type_get_attr",
  "MPI_Type_get_contents",
  "MPI_Type_get_envelope",
  "MPI_Type_get_extent",
  "MPI_Type_get_name",
  "MPI_Type_get_true_extent",
  "MPI_Type_hindexed",
  "MPI_Type_hvector",
  "MPI_Type_indexed",
  "MPI_Type_lb",
  "MPI_Type_match_size",
  "MPI_Type_set_attr",
  "MPI_Type_set_name",
  "MPI_Type_size",
  "MPI_Type_struct",
  "MPI_Type_ub",
  "MPI_Type_vector",
  "MPI_Unpack",
  "MPI_Unpack_external",
  "MPI_Unpublish_name",
  "MPI_Wait",
  "MPI_Waitall",
  "MPI_Waitany",
  "MPI_Waitsome",
  "MPI_Win_c2f",
  "MPI_Win_call_errhandler",
  "MPI_Win_complete",
  "MPI_Win_create",
  "MPI_Win_create_errhandler",
  "MPI_Win_create_keyval",
  "MPI_Win_delete_attr",
  "MPI_Win_f2c",
  "MPI_Win_fence",
  "MPI_Win_free",
  "MPI_Win_free_keyval",
  "MPI_Win_get_attr",
  "MPI_Win_get_errhandler",
  "MPI_Win_get_group",
  "MPI_Win_get_name",
  "MPI_Win_lock",
  "MPI_Win_post",
  "MPI_Win_set_attr",
  "MPI_Win_set_errhandler",
  "MPI_Win_set_name",
  "MPI_Win_start",
  "MPI_Win_test",
  "MPI_Win_unlock",
  "MPI_Win_wait",
  "MPI_Wtick",
  "MPI_Wtime"
};

/** Mapping of string keys to enabling group IDs
 * @note The keys need to be alphabetically sorted, to enable binary
 *       search on the data structure.
 */
static const EpkMPItype epk_mpi_enable_groups[] = {
  { "ALL",       LWM2_MPI_GROUP_ALL              },
  { "CG",        LWM2_MPI_GROUP_CG               },
  { "COLL",      LWM2_MPI_GROUP_COLL             },
  { "DEFAULT",   LWM2_MPI_GROUP_DEFAULT          },
  { "ENV",       LWM2_MPI_GROUP_ENV              },
  { "ERR",       LWM2_MPI_GROUP_ERR              },
  { "EXT",       LWM2_MPI_GROUP_EXT              },
  { "IO",        LWM2_MPI_GROUP_IO               },
  { "MISC",      LWM2_MPI_GROUP_MISC             },
  { "P2P",       LWM2_MPI_GROUP_P2P              },
  { "PERF",      LWM2_MPI_GROUP_PERF             },
  { "RMA",       LWM2_MPI_GROUP_RMA              },
  { "SPAWN",     LWM2_MPI_GROUP_SPAWN            },
  { "TOPO",      LWM2_MPI_GROUP_TOPO             },
  { "TYPE",      LWM2_MPI_GROUP_TYPE             },
  { "XNONBLOCK", LWM2_MPI_GROUP_XNONBLOCK        }
};
