/****************************************************************************
**  Wrapper originally built for:										
**  **
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  LWM2
**  **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2010                                                     **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/**
 * @file  epk_mpiwrap_env.c
 *
 * @brief C interface wrappers for environmental management
 */

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#include "../lwm2-mpi-config.h"
#include "lwm2_mpireg.h"

/*
   CS
 */
#define elg_error_msg(...)

/**
 * @name C wrappers
 * @{
 */

/**
 * Measurement wrapper for MPI_Init
 * @note Manually adapted wrapper
 * @note C interface
 * @note Introduced with MPI 1.0
 * @ingroup env
 */
int MPI_Init(int* argc, char*** argv)
{
  int            event_gen_active = 0; /* init is deferred to later */
  int            return_val, numprocs, i;
  unsigned char* grpv;
  int            fflag;

  event_gen_active = IS_EVENT_GEN_ON_FOR(ENV);

  region_information_t _info;
  if (event_gen_active)
  {
    _info = lwm2_mpi_enter(LWM2_MPI_INIT, LWM2_MPI_GROUP_ENV);
  }

  return_val = PMPI_Init(argc, argv);

  if ((PMPI_Finalized(&fflag) == MPI_SUCCESS) && (fflag == 0))
  {
    /* Initialize lwm2 */
    lwm2_mpi_init();
  }

  if (event_gen_active)
  {
    lwm2_mpi_leave(&_info);
  }

  return return_val;
} /* MPI_Init */

#if defined(HAS_MPI_INIT_THREAD)
/**
 * Measurement wrapper for MPI_Init_thread, the thread-capable
 * alternative to MPI_Init.
 * @note Manually adapted wrapper
 * @note C interface
 * @note Introduced with MPI 2.0
 * @ingroup env
 */
int MPI_Init_thread(int* argc, char*** argv, int required, int* provided)
{
  int            event_gen_active = 0;
  int            return_val, numprocs, i;
  unsigned char* grpv;
  int            fflag;
  int            already_opened;

  /* open output files and return whether the files had been opened
   * elsewhere in the measurement */
  /*already_opened = esd_open();*/
  /* esd_open calls epk_mpi_register(), so we have to defer the setting
   * of this test to now */
  event_gen_active = IS_EVENT_GEN_ON_FOR(ENV);

  region_information_t _info;
  if (event_gen_active)
  {
    _info = lwm2_mpi_enter(LWM2_MPI_INIT_THREAD, LWM2_MPI_GROUP_ENV);
  }

  return_val = PMPI_Init_thread(argc, argv, required, provided);

  /* XXXX should only continue if MPI initialised OK! */

  if ((return_val == MPI_SUCCESS) && (*provided > MPI_THREAD_FUNNELED))
  {
    /*elg_warning(
       "MPI environment initialized with level exceeding
          MPI_THREAD_FUNNELED!");*/
    /* XXXX continue even though not supported by analysis */
  }

  if ((PMPI_Finalized(&fflag) == MPI_SUCCESS) && (fflag == 0))
  {
	  lwm2_mpi_init();
  }

  if (event_gen_active)
  {
    lwm2_mpi_leave(&_info);
  }

  return return_val;
} /* MPI_Init_thread */

#else
int MPI_Init_thread(int*    argc,
                    char*** argv,
                    int     required,
                    int*    provided)
{
  elg_error_msg(
    "Abort: MPI_Init_thread wrapper was not built correctly. Please reinstall Scalasca with enabled MPI_Init_thread support.");
}

#endif

/**
 * Measurement wrapper for MPI_Finalize
 * @note Manually adapted wrapper
 * @note C interface
 * @note Introduced with MPI 2.0
 * @ingroup env
 */
int MPI_Finalize()
{
  int return_val;

  /* finalize MPI event handling */
  lwm2_mpi_finalize();

  /* fake finalization, so that MPI can be used during finalization */
//  return_val = PMPI_Barrier(MPI_COMM_WORLD);

  return_val = PMPI_Finalize();

  return return_val;
}

#if defined(HAS_MPI_IS_THREAD_MAIN) && !defined(NO_MPI_EXTRA) \
  && !defined(NO_MPI_ENV)
/**
 * Measurement wrapper for MPI_Is_thread_main
 * @note Auto-generated by wrapgen from template: std.w
 * @note C interface
 * @note Introduced with MPI 2.0
 * @ingroup env
 */
int MPI_Is_thread_main(int* flag)
{
  int                  return_val;

  region_information_t _info = lwm2_mpi_enter(LWM2_MPI_IS_THREAD_MAIN,
                                              LWM2_MPI_GROUP_ENV);

  return_val = PMPI_Is_thread_main(flag);

  lwm2_mpi_leave(&_info);

  return return_val;
}

#endif
#if defined(HAS_MPI_QUERY_THREAD) && !defined(NO_MPI_EXTRA) \
  && !defined(NO_MPI_ENV)
/**
 * Measurement wrapper for MPI_Query_thread
 * @note Auto-generated by wrapgen from template: std.w
 * @note C interface
 * @note Introduced with MPI 2.0
 * @ingroup env
 */
int MPI_Query_thread(int* provided)
{
  int                  return_val;

  region_information_t _info = lwm2_mpi_enter(LWM2_MPI_QUERY_THREAD,
                                              LWM2_MPI_GROUP_ENV);

  return_val = PMPI_Query_thread(provided);

  lwm2_mpi_leave(&_info);

  return return_val;
}

#endif

#if defined(HAS_MPI_FINALIZED) && !defined(NO_MPI_EXTRA) \
  && !defined(NO_MPI_ENV)
/**
 * Measurement wrapper for MPI_Finalized
 * @note Auto-generated by wrapgen from template: std.w
 * @note C interface
 * @note Introduced with MPI 2.0
 * @ingroup env
 */
int MPI_Finalized(int* flag)
{
  int                  return_val;

  region_information_t _info = lwm2_mpi_enter(LWM2_MPI_FINALIZED,
                                              LWM2_MPI_GROUP_ENV);

  return_val = PMPI_Finalized(flag);

  lwm2_mpi_leave(&_info);

  return return_val;
}

#endif
#if defined(HAS_MPI_INITIALIZED) && !defined(NO_MPI_EXTRA) \
  && !defined(NO_MPI_ENV)
/**
 * Measurement wrapper for MPI_Initialized
 * @note Auto-generated by wrapgen from template: std.w
 * @note C interface
 * @note Introduced with MPI 1.0
 * @ingroup env
 */
int MPI_Initialized(int* flag)
{
  int                  return_val;

  region_information_t _info = lwm2_mpi_enter(LWM2_MPI_INITIALIZED,
                                              LWM2_MPI_GROUP_ENV);

  return_val = PMPI_Initialized(flag);

  lwm2_mpi_leave(&_info);

  return return_val;
}

#endif
/**
 * @}
 */
