/****************************************************************************
**  Copyright (c) 2011-2012                                                **
**  German Research School for Simulation Sciences GmbH					   **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "util/lwm2-log.h"

#include "lwm2-memory.h"
#include "lwm2-storage.h"

/*
 * time-slice's linked-list
 */
char *lwm2_time_slice_buffer_head = NULL;
char *lwm2_time_slice_buffer_current = NULL;

/*
 * temp buffer's linked list
 */
char* lwm2_temp_buffer_head = NULL;
char* lwm2_temp_buffer_current = NULL;

char* lwm2_end_temp_buffer_head = NULL;
char* lwm2_end_temp_buffer_current = NULL;

/*
 * thread buffer, fix size with respect to group count. Index specifies the current group
 */
char *lwm2_thread_buffer[THREAD_GROUP_COUNT];
int lwm2_thread_buffer_index = 0;

// indicate if storage initialized or not
long lwm2_thread_storage_unit = 0;

/*
 * Initialize buffers
 */
int lwm2_memory_init(void)
{	
	lwm2_memory_init_buffer();

	return LWM2_SUCCESS;
}

/*
 * Free up the memory for time-slices, temp buffers and threads.
 */
void lwm2_memory_destroy(void)
{
	
	free(lwm2_time_slice_buffer_head);
	lwm2_time_slice_buffer_current = lwm2_time_slice_buffer_head = NULL;
	
	free(lwm2_temp_buffer_head);
	lwm2_temp_buffer_current = lwm2_temp_buffer_head = NULL;
	
	int i;
	for(i = 0; i < THREAD_GROUP_COUNT; i++)
	{
		free(lwm2_thread_buffer[i]);
		lwm2_thread_buffer[i] = NULL;
	}
}

/*
 * Initialize all the buffers
 */
void lwm2_memory_init_buffer()
{
	//Initialize the time-slice 
	lwm2_time_slice_buffer_head = malloc(TIME_SLICE_SIZE * TIME_SLICE_COUNT);
	if(lwm2_time_slice_buffer_head == NULL)
	{
		LWM2LOG(0, "Can't allocate memory for time-slices. Aborting!\n");
		exit(-1);
	}
	lwm2_time_slice_buffer_current = lwm2_time_slice_buffer_head;
	
	//Initialize temp buffer
	lwm2_temp_buffer_head = malloc(TEMP_MEM_CHUNK);
	if(lwm2_temp_buffer_head == NULL)
	{
		LWM2LOG(0, "Can't allocate memory for temp buffer. Aborting\n");
		exit(-1);
	}
	lwm2_temp_buffer_current = lwm2_temp_buffer_head;
	
	//Initialize storage for threads. Get the Page size of the system and assign 
	//each thread a page
	lwm2_thread_storage_unit = sysconf(_SC_PAGE_SIZE);
	
	if(lwm2_thread_storage_unit == -1)
	{
		lwm2_thread_storage_unit = THREAD_SLICE_DEFAULT_SIZE;
	}
	
	int i;
	for(i = 0; i < THREAD_GROUP_COUNT; i++)
	{
		lwm2_thread_buffer[i] = malloc(THREAD_GROUP_SIZE * lwm2_thread_storage_unit);
		
		if(lwm2_thread_buffer[i] == NULL)
		{
			LWM2LOG(0, "Can't allocate memory for thread storage. Aborting!\n");
			exit(-1);
		}
	}
}

/*
 * Allocates the specified bytes from the time-slice buffer and returns the pointer.
 * This function should not be called directly, but the function,
 * lwm2_memory_new_time_slice_storage() should be called which allocates
 * the buffer with size of the defined time-slice
 */
void* lwm2_memory_new_time_slice_storage_bytes(int bytes)
{
	if(lwm2_time_slice_buffer_current + bytes >= lwm2_time_slice_buffer_head + (TIME_SLICE_SIZE * TIME_SLICE_COUNT))
	{
		LWM2LOG(0, "Time-slice buffer overrun! Aborting\n");
		exit(-1);
	}
	void* buffer = (void*) lwm2_time_slice_buffer_current;
	lwm2_time_slice_buffer_current += bytes;
	
	return buffer;
}

/*
 * Allocate a buffer of specified bytes from the temp buffer 
 */
void* lwm2_memory_new_temp_buffer_bytes(int bytes)
{
	if(lwm2_temp_buffer_current + bytes >= lwm2_temp_buffer_head + TEMP_MEM_CHUNK)
	{
		LWM2LOG(0, "Temp buffer overrun! Aborting\n");
		exit(-1);
	}
	void* buffer = (void*) lwm2_temp_buffer_current;
	lwm2_temp_buffer_current += bytes;
	
	return buffer;
}

/*
 * Free the temp buffer of any allocated bytes.
 * Note: This does not give up the memory allocated. 
 * This only allows to reuse the memory
 */
void lwm2_memory_free_temp_buffer()
{
	lwm2_temp_buffer_current = lwm2_temp_buffer_head;
}

/*
 * Allocate a buffer of specified bytes from the end temp buffer.
 * The end temp buffer should be used after the profiled application
 * has ended. It allocates a larger chunk of memory compared to temp. 
 */
void* lwm2_memory_new_end_temp_buffer_bytes(int bytes)
{
	if(lwm2_end_temp_buffer_head == NULL)
	{
		lwm2_end_temp_buffer_head = malloc(END_TEMP_MEM_CHUNK);
		if(lwm2_end_temp_buffer_head == NULL)
		{
			LWM2LOG(0, "Can't allocate memory for end temp buffer. Aborting\n");
			exit(-1);
		}
		lwm2_end_temp_buffer_current = lwm2_end_temp_buffer_head;
	}
	
	
	if(lwm2_end_temp_buffer_current + bytes >= lwm2_end_temp_buffer_head + END_TEMP_MEM_CHUNK)
	{
		LWM2LOG(0, "Temp buffer overrun! Aborting\n");
		exit(-1);
	}
	void* buffer = (void*) lwm2_end_temp_buffer_current;
	lwm2_end_temp_buffer_current += bytes;

	return buffer;
}

/*
 * Free the end temp buffer of any allocated bytes.
 * Note: This does not give up the memory allocated. 
 * This only allows to reuse the memory
 */
void lwm2_memory_free_end_temp_buffer()
{
	lwm2_end_temp_buffer_current = lwm2_end_temp_buffer_head;
}

/*
 * Allocates a buffer of size 1 time-slice
 */
unsigned long long* lwm2_memory_new_time_slice_storage()
{
	void *buffer = lwm2_memory_new_time_slice_storage_bytes(TIME_SLICE_SIZE);
	return (unsigned long long*) buffer;
}

/*
 * Allocate a new buffer for a thread
 */
void* lwm2_new_thread_storage()
{
	if(lwm2_thread_buffer_index >= (THREAD_GROUP_SIZE * THREAD_GROUP_COUNT))
	{
		LWM2LOG(0, "Thread buffer overrun!. Aborting\n");
		exit(-1);
	}
	
	//Find out the thread group and the index in the group for the overall thread index
	int thread_group = lwm2_thread_buffer_index / THREAD_GROUP_SIZE;
	int thread_index = lwm2_thread_buffer_index - thread_group * THREAD_GROUP_SIZE;
	
	//Get the pointer to that memory location
	char* group_buffer = lwm2_thread_buffer[thread_group];
	void* thread_buffer = (void*) (group_buffer + (thread_index * lwm2_thread_storage_unit));
	
	lwm2_thread_buffer_index++;
	
	return thread_buffer;
}
