/****************************************************************************
**  Copyright (c) 2011-2012                                                **
**  German Research School for Simulation Sciences GmbH					   **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#define _GNU_SOURCE
#include <dlfcn.h>

#ifndef LWM2_DEV_NOLIBINIT
/* Library initialization */
#if defined(__GNU_C) || defined(__INTEL_COMPILER) || defined(__GNUC__) || defined(__IBMC__)
/* GNU - Library init/end functions */ 
void __attribute__ ((constructor))
lwm2_init(void);

void __attribute__ ((destructor))
lwm2_end(void);
#else
/* Not GNU C */
#error("No constructor method available for this compiler!")
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <ucontext.h>
#include <execinfo.h>
#include <math.h>
#include <string.h>
#include <time.h>

/* Threading API */
#include <pthread.h>

/* LWM2 headers */
#include "lwm2-signal.h"
#include "lwm2-memory.h"

#include "util/lwm2-log.h"
#include "util/util_time_measurement.h"
#include "util/lwm2_ignore_reader.h"

#include "lwm2-system.h"

#include "lwm2-error-handling.h"

#include "lwm2-storage.h"

#include "modules/output/lwm2-output-socket.h"

#include "external/lwm2file/include/lwm2-file-reader.h"
#include "modules/output/lwm2-output-file.h"
#include "modules/output/lwm2-output-summary.h"

#include "lwm2-pulse-thread.h"

/* Include modules */
/*************************************************************************/
/* Measurement modules */

/* Default module */
#include "modules/measurements/lwm2-measurement-default.h"

#include "modules/measurements/lwm2-measurement-mpi.h"

#include "adapter/lwm2-mpi-adapter.h"
#include "adapter/lwm2-omp-adapter.h"


/* OpenMP Measurement Module */
#include "modules/measurements/lwm2-measurement-omp.h"



/* pthreads Measurement Module */
#include "modules/measurements/lwm2-measurement-pthread.h"



/* IO Measurement Module */
#include "modules/measurements/lwm2-measurement-io.h"


/* HWC Measurement Module */
#include "modules/measurements/lwm2-measurement-hwcounter.h"

#include "modules/measurements/lwm2-measurement-cuda.h"

//length of a time-slice
int lwm2_time_slice_duration = 10; 

//a flag to indicate LWM2 was called for a system service
int lwm2_sys_service_active = 0;

/* Constants */
#define DEFAULT_SAMPLING_RATE 100
long lwm2_sampling_rate = 100; /* in hz */

/* Calculated */
long lwm2_timer_interval;

/* Flags depicting status information */
unsigned int lwm2_initialized = 0;

/* Fields to enable subtraction of wrapper time from signals */
volatile sig_atomic_t lwm2_skip_signal = 0; /* Skip a signal */

//To indicate the application has ended
int lwm2_app_ended = 0;

//To store job id 
int lwm2_job_id = -1;

//Store application start and end time
timestamp_t lwm2_first_time_slice = 0;
timestamp_t lwm2_start_time;
timestamp_t lwm2_end_time;
time_t lwm2_long_start_time = 0;

/**
 * SIGPROF Timer signal handler
 */
static void sigprof_handler(int sig_nr, siginfo_t* info, void *context)
{
	//if a signal handler is already in progress, skip the signal
	if (lwm2_storage_get_flag_sighandler() > 0)
	{
		return;
	}

	//set the signal handler flag
	lwm2_storage_set_flag_sighandler();

	/* Skip signal, mostly in case of error */
	if (lwm2_skip_signal != 0)
	{
		/* Reset signal handler flag */
		lwm2_storage_reset_flag_sighandler();

		return;
	}

	lwm2_measurement_handle_signal_default();

	lwm2_measurement_handle_signal_mpi();

	lwm2_measurement_handle_signal_cuda();

	lwm2_measurement_handle_signal_fileio();

	lwm2_measurement_handle_signal_omp();

	//Reset the signal handler flag
	lwm2_storage_reset_flag_sighandler();
}

//This function is called my a timer, when it is required to read the hardware counters
//This function should be called in the context of each thread, for which the HWC have to
//measured. Otherwise, the HWCs of the calling thread will be measured, if the thread
//is registered with PAPI
void read_thread_hwc(int sig_nr, siginfo_t* info, void *context)
{
#ifdef LWM2_PAPI_THREAD_ENABLED
	int thread_index = info->si_value.sival_int;
	struct lwm2_hwc_storage* hwc_storage = lwm2_thread_storage_table[thread_index].hwc_storage;
	lwm2_measurement_hwc_papi_read_reset_events(hwc_storage);
	lwm2_thread_storage_table[thread_index].did_read_hwc = 1;
#endif
}

/*
 * Called by pulse thread when a time-slice ends.
 */
void lwm2_signal_update_time_slice()
{
	if(lwm2_first_time_slice == 0)
	{
		lwm2_first_time_slice = util_get_timestamp();
	}
	lwm2_last_time_slice_time = lwm2_util_get_long_time();
	if(lwm2_first_time_slice_time == 0)
	{
		lwm2_first_time_slice_time = lwm2_last_time_slice_time;
	}
	
#ifdef LWM2_MEASUREMENT_HWCOUNTER
	//Read the hardware counters before updating the time-slice
#ifndef LWM2_PAPI_THREAD_ENABLED
	lwm2_measurement_hwc_papi_read_reset_events(&lwm2_proc_hwc_storage);
#endif

#endif
	
	lwm2_measurement_read_cuda_counters();
	
	lwm2_storage_update_time_slice();
	
	lwm2_storage_reset();
}


/**
 * Enable SIGPROF timer
 */
void timer_enable(void)
{
	static struct itimerval timer;

	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = lwm2_timer_interval;
	timer.it_value = timer.it_interval;

	if (setitimer(ITIMER_PROF, &timer, NULL) != 0)
	{
		LWM2LOG(1, "Error: Timer could not be initialized. Sampling will be disabled\n");
	}
}

/**
 * Disables the SIGPROF timer
 */
void lwm2_timer_disable(void)
{
	static struct itimerval timer;

	getitimer(ITIMER_PROF, &timer);

	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = 0;
	timer.it_value = timer.it_interval;

	setitimer(ITIMER_PROF, &timer, NULL);
}

/*
 * Fetch the Job ID from the environment variable. Currently hardcoded for JUROPA.
 */
void lwm2_get_job_id_juropa()
{
	//JUROPA job id format: 1787473.jj28b01
	char *job_id = getenv("PBS_JOBID");
	if(job_id == NULL)
	{
		lwm2_job_id = 0;
		return;
	}
	
	//Get the index for the '.'
	char *dot = strchr(job_id, '.');

	//Get the length
	int job_id_len = job_id - dot;
	
	//Can this really happen?
	if(job_id_len < 0)
	{
		job_id_len = job_id_len * -1;
	}

	//store the job id as string
	char specific_job_id[32];
	strncpy(specific_job_id, job_id, job_id_len);
	specific_job_id[job_id_len] = 0;
	
	//now convert it into an integer
	lwm2_job_id = atoi(specific_job_id);
}

/*
 * A generic job id function. The environment variable' name containing the job id
 * must be specified by the  variable LWM2_JOBID_VAR. It is assumed that the first
 * integer part of the job id variable will contain the actual job id.
 */
void lwm2_get_generic_job_id()
{
	char* job_id_identifier = getenv("LWM2_JOBID_VAR");
	if(job_id_identifier == NULL)
	{
		lwm2_job_id = 0;
		return;
	}
	char* job_id = getenv(job_id_identifier);
	
	if(job_id == NULL)
	{
		lwm2_job_id = 0;
		return;
	}
	
	char int_job_id[8];
	
	sscanf(job_id, "%[0123456789]", int_job_id);
	sscanf(int_job_id, "%d", &lwm2_job_id);
}

/*
 * Calls the appropriate job id function, depending upon the definition
 */
void lwm2_get_job_id()
{
#ifdef MACHINE_JUROPA
	lwm2_get_job_id_juropa();
#elif MACHINE_JUDGE
	lwm2_get_job_id_juropa();
#else
	lwm2_get_generic_job_id();
#endif
}

/*
 * Gets the sampling rate from the environment variable
 */
long lwm2_get_sampling_rate()
{
	char* sampling_rate = getenv("LWM2_SAMPLING_RATE");
	long rate = DEFAULT_SAMPLING_RATE;
	if(sampling_rate != NULL)
	{
		long r = atol(sampling_rate);
		if(r != 0)
		{
			rate = r;
		}
	}
	return rate;
}

int lwm2_get_time_slice_length()
{
	char* time_slice_length = getenv("LWM2_TIME_SLICE_LENGTH");
	int length = lwm2_time_slice_duration;
	if(time_slice_length != NULL)
	{
		int l = atoi(time_slice_length);
		if(l != 0)
		{
			length = l;
		}
	}
	return length;
}

/**
 * The Init function called by the runtime whenever an application starts, and LWM2 is preloaded.
 */
void lwm2_init(void)
{
	char bin_name[1024];
	
	//calculate the sampling rate first
	lwm2_sampling_rate = lwm2_get_sampling_rate();
	lwm2_time_slice_duration = lwm2_get_time_slice_length();

	//Calculate the timer interval from the sampling rate
	lwm2_timer_interval = 1000000 / lwm2_sampling_rate; /* Microseconds */

	//Check for multiple initialization. Can this happen?
	if (lwm2_initialized > 0)
	{
		/* Library has already been initialized */
		LWM2LOG(4,	"Initialization called even though lib is already initialized. \n");

		return;
	}
	
	lwm2_read_ignore_file();

	/* Retrieve binary name */
	lwm2_system_binary_name(bin_name);
	
	//reads the list of ignored names from a file specified by environment variables
	if(lwm2_should_ignore(bin_name))
	{
		LWM2LOG(4, "lwm2_init called for system services. Ignoring...\n");

		lwm2_sys_service_active = 1;

		return;
	}
	else
	{
		lwm2_sys_service_active = 0;
		lwm2_start_time = util_get_timestamp();
		lwm2_long_start_time = lwm2_util_get_long_time();
	}
	
	//setup I/O by dynamic function interposing
	lwm2_measurement_setup_io();

	//setup pthread by dynamic function interposing
	lwm2_measurement_setup_pthread();

	//get the job id from the environment variable
	lwm2_get_job_id();

	/* NOTE: Error subsystem disabled as it is not tested yet! */
//   if (LWM2_SUCCESS != lwm2_error_init())
//   {
//     
//
//      LWM2TRACE(1, "Error: Cannot initialize error subsystem \n");
//   }

	/* Setup memory management */
	lwm2_memory_init();

	//setup storage of metrics
	lwm2_storage_init();

	/* Install signal handler for SIGPROF event */
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	sa.sa_sigaction = sigprof_handler;
	sa.sa_flags = SA_RESTART | SA_SIGINFO;

	sigfillset(&sa.sa_mask);

	if(sigaction(SIGPROF, &sa, NULL) != 0)
	{
		LWM2LOG(1, "Error while setting up the signal handler. Sampling will be disabled\n");
	}
	
#ifdef LWM2_PAPI_THREAD_ENABLED
	struct sigaction tsa;
	memset(&tsa, 0, sizeof(tsa));
	tsa.sa_sigaction = read_thread_hwc;
	tsa.sa_flags = SA_RESTART | SA_SIGINFO;

	sigfillset(&tsa.sa_mask);

	if(sigaction(SIGALRM, &tsa, NULL) != 0)
	{
		LWM2LOG(1, "Error while setting up the signal handler for reading thread HW counters\n");
	}
#endif

	/* Init measurement modules */
	lwm2_measurement_setup_default();

	lwm2_measurement_setup_mpi();

	lwm2_measurement_setup_omp();

	lwm2_measurement_setup_hwcounter();

	lwm2_measurement_setup_cuda();

	//get the output file name
	lwm2_output_file_get_l2f_filename();

	//enable timer, if not disabled during compilation
#ifndef LWM2_DISABLE_SAMPLING
	timer_enable();
#endif
	
	lwm2_thread_start_pulse_thread();

	/* Initialization  has now been finished */
	lwm2_initialized = 1;
}

/**
 * End function called by the runtime whenever an application ends, for which LWM2 has been preloaded
 */
void lwm2_end(void)
{
	if(lwm2_app_ended == 1)
	{
		return;
	}

	//No need to do anything for a system service
	if(lwm2_sys_service_active != 0)
	{
		LWM2LOG(2, "lwm2_end called for system services. Ignoring...\n");
		return;
	}

	/* Prohibit interruption of end routine by signal handler */
	lwm2_skip_signal = 1;
	lwm2_timer_disable();

	/* End all measurement modules */
	lwm2_measurement_teardown_default();

	lwm2_measurement_teardown_mpi();

	lwm2_measurement_teardown_pthread();

	lwm2_measurement_teardown_omp();

#ifdef LWM2_MEASUREMENT_HWCOUNTER
	if(lwm2_mpi_app == 0)
	{
		lwm2_measurement_hwc_papi_read_reset_events(&lwm2_proc_hwc_storage);
	}
#endif
	
	lwm2_measurement_teardown_hwcounter();


	lwm2_app_ended = 1;

	//If its an MPI application and this is the root process.
	if ((lwm2_mpi_rank == 0)) 
	{

		lwm2_end_time = util_get_timestamp();
		lwm2_output_summary_print_summary();
	}

	//If this is not an MPI application
	if((lwm2_mpi_app == 0)) 
	{
		lwm2_end_time = util_get_timestamp();
		lwm2_signal_update_time_slice();
		lwm2_storage_merge_threads_sum_data();
		lwm2_output_file_write_non_mpi_app_file();
		lwm2_output_summary_print_summary_non_mpi();
	}

	//Disconnect from ClustrX
	lwm2_socket_disconnect();


	/* Cleanup */
	lwm2_memory_destroy();

}

