/****************************************************************************
**  Copyright (c) 2011-2012                                                **
**  German Research School for Simulation Sciences GmbH					   **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "lwm2-storage.h"
#include "util/lwm2-log.h"
#include "util/util_time_measurement.h"
#include "lwm2-memory.h"
#include "modules/output/lwm2-output-storage.h"
#include "modules/output/lwm2-output-file.h"
#include "modules/output/lwm2-output-socket.h"
#include "adapter/lwm2-mpi-adapter.h"
#include "adapter/lwm2-pthread-adapter.h"
#include "adapter/lwm2-hwc-adapter.h"

#include "adapter/lwm2-cuda-adapter.h"
#include "modules/measurements/lwm2-measurement-omp.h"
#include "lwm2-signal.h"
#include "adapter/hwc/lwm2-hwc-papi.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <mpi.h>
#include <pthread.h>
#include <unistd.h>

double lwm2_root_first_time_slice;

long lwm2_time_slice_index = 1;

//Names of the metrics based on signals
char* lwm2_signal_metric_names[] = {
		"Total Signals",
		"MPI Signals",
		"File I/O Signals",
		"MPI I/O Signals",
		"CUDA Signals",
		"CUDA to device signals",
		"CUDA to host signals",
		"MPI P2P Signals" ,
		"MPI Coll Signals"
};


#ifndef LWM2_HWC_EXTERN_DEF

//Names of metrics for HWCs
//These are the default names, which are used when an external HWC list is not provided
char* lwm2_hwc_metric_names[] = {
		"Cycles",
		"FL Operations",
		"FLOPS",
		"Instructions",
		"L1 Data Accesses",
		"L1 Data misses"
};
#endif

//Rest of the metrics, which are captured from wrappers
char* lwm2_tsliced_metric_names[] = {
		"P2P bytes sent",
		"P2P bytes recv",
		"Coll bytes sent",
		"Coll bytes recv",
		"P2P msgs sent",
		"P2P msgs recv",
		"Coll call count",
		"File I/O bytes read",
		"File I/O bytes written",
		"CUDA bytes to device",
		"CUDA bytes to host",
		"CUDA msgs to device",
		"CUDA msgs to host",
		"CUDA Instructions",
		"CUDA divergent branches",
		"Max. no. of threads"
};

//storage used to store metrics values for the whole execution
unsigned long long lwm2_metric_total[LWM2_TOTAL_METRIC_COUNT];

//used to manage time-slice linked list
struct metric_node* list_head = NULL;
struct metric_node* current_node = NULL;

//used to store thread specific data
struct lwm2_thread_storage lwm2_thread_storage_table[MAX_THREADS];
pthread_rwlock_t lwm2_thread_storage_table_lock;

//lock for thread table
pthread_rwlock_t lwm2_process_storage_lock;

//page size of the system
long lwm2_system_page_size;

//some specific metrics to be stored for summary
unsigned long long lwm2_mpi_comm_data[COMM_DATA_TOTAL];

/*
 * Get the process or thread storage for storing complete execution metrics
 * If a thread is registered, the value is stored in per thread storage
 * otherwise it is stored in the central storage of a process
 */
unsigned long long* lwm2_storage_get_my_storage_sum()
{
	unsigned long long* thread_storage = NULL;
	struct lwm2_thread_storage* pthread_storage = pthread_getspecific(lwm2_pthread_data_key);
	//This means the thread hasn't registered anything, meaning no thread storage.
	if(pthread_storage == NULL)
	{
		thread_storage = lwm2_metric_total;
	}
	else
	{
		thread_storage = pthread_storage->thread_data_sum;
	}
	return thread_storage;
}

//Get the process or thread storage for storing time-slice metrics
//If a thread is registered, the value is stored in per thread storage
//otherwise it is stored in the central storage of a process
unsigned long long* lwm2_storage_get_my_storage_slice(pthread_rwlock_t* thread_lock)
{
	unsigned long long* thread_storage = NULL;
	thread_lock = NULL;
	struct lwm2_thread_storage* pthread_storage = pthread_getspecific(lwm2_pthread_data_key);
	//This means the thread hasn't registered anything, meaning no thread storage.
	if(pthread_storage == NULL)
	{
		thread_storage = current_node->metrics_data;
	}
	else
	{
		//Check which buffer to return for the thread storage.
		//Each consecutive time-slice is stored in an alternating buffer
		
		//Need to lock the thread table before reading the value. Read lock is used
		pthread_rwlock_rdlock(&lwm2_thread_storage_table_lock);
		if(pthread_storage->thread_data_slice_index == 0)
		{
			thread_storage = pthread_storage->thread_data_slice_1;
		}
		else
		{
			thread_storage = pthread_storage->thread_data_slice_2;
		}
		thread_lock = pthread_storage->thread_lock;
		pthread_rwlock_unlock(&lwm2_thread_storage_table_lock);
	}
	return thread_storage;
}

/*
 * Get the storage used to store HWCs. If a per thread is present, use that,
 * otherwise used the process storage
 * If a thread is registered, the value is stored in per thread storage
 * otherwise it is stored in the central storage of a process
 * Note: returning void* because the structure is defined only when HWC are measured
 */
void* lwm2_storage_get_my_hwc_storage()
{
#ifdef LWM2_MEASUREMENT_HWCOUNTER
	
#ifdef LWM2_PAPI_THREAD_ENABLED
	struct lwm2_thread_storage* pthread_storage = pthread_getspecific(lwm2_pthread_data_key);
	if(pthread_storage == NULL)
	{
		return (void*) &lwm2_proc_hwc_storage;
	}
	else
	{
		return (void*) pthread_storage->hwc_storage;
	}
#endif
	
#endif
	return (void*) NULL;
}

/*
 * Release the thread specific storage used to store HWCs.
 * Note, memory is not released, but the reference to it is removed.
 * Deallocation occurs in memory module
 */
void lwm2_storage_reset_my_hwc_storage()
{
#ifdef LWM2_MEASUREMENT_HWCOUNTER
	
#ifdef LWM2_PAPI_THREAD_ENABLED
	struct lwm2_thread_storage* pthread_storage = pthread_getspecific(lwm2_pthread_data_key);
	if(pthread_storage != NULL)
	{
		pthread_storage->hwc_storage = NULL;
	}
#endif
	
#endif
}

//Check if the thread is registered
int lwm2_storage_am_i_a_registered_thread()
{
	struct lwm2_thread_storage* pthread_storage = pthread_getspecific(lwm2_pthread_data_key);
	if(pthread_storage == NULL)
	{
		return 0;
	}
	return 1;
}

//Get the thread specific flag storage
//If thread is not registered, return the process flag storage
char* lwm2_storage_get_flag_storage()
{
	struct lwm2_thread_storage* thread_storage = pthread_getspecific(lwm2_pthread_data_key);
	char* flag_storage = NULL;
	if(thread_storage != NULL)
	{
		flag_storage = thread_storage->my_flags;
	}
	else
	{
		flag_storage = lwm2_all_flags;
	}
	
	return flag_storage;
}

//Switch between the thread storages. This happens at the end of each time-slice
void lwm2_storage_change_thread_storage()
{
	int i;

	//Use locks before modifying the index value
	pthread_rwlock_wrlock(&lwm2_thread_storage_table_lock);
	for(i = 0; i < lwm2_omp_thread_count; i++)
	{
		lwm2_thread_storage_table[i].thread_data_slice_index = (lwm2_thread_storage_table[i].thread_data_slice_index + 1) % 2;
	}
	pthread_rwlock_unlock(&lwm2_thread_storage_table_lock);
}

//Merge the thread data with the process data, at the end of time-slice
void lwm2_storage_merge_threads_time_slice_data()
{
	int i;
	unsigned long long* thread_storage;
	pthread_rwlock_t* thread_lock;
	
	pthread_rwlock_wrlock(&lwm2_thread_storage_table_lock);
	for(i = 0; i < lwm2_omp_thread_count; i++)
	{
		if(lwm2_thread_storage_table[i].thread_data_slice_index == 0)
		{
			thread_storage = lwm2_thread_storage_table[i].thread_data_slice_2;
		}
		else
		{
			thread_storage = lwm2_thread_storage_table[i].thread_data_slice_1;
		}
		thread_lock = lwm2_thread_storage_table[i].thread_lock;
		lwm2_storage_merge_thread_time_slice_data(thread_storage, current_node->metrics_data, thread_lock);
	}
	pthread_rwlock_unlock(&lwm2_thread_storage_table_lock);
}

//Merges the thread data into the process data. The buffers are resolved and passed
void lwm2_storage_merge_thread_time_slice_data(unsigned long long* thread_data, unsigned long long* process_data, pthread_rwlock_t* thread_lock)
{
	int i;
	
	if(thread_data == NULL)
	{
		LWM2LOG(2, "[%d] The thread storage is NULL!!\n", lwm2_mpi_rank);
		return;
	}
	
	if(process_data == NULL)
	{
		return;
	}
	
	//Lock before accessing the data
//	pthread_rwlock_wrlock(&lwm2_process_storage_lock);
	pthread_rwlock_wrlock(thread_lock);
	for( i = 0; i < LWM2_WRAPPER_COUNT; i++)
	{
		process_data[i] += thread_data[i];
		thread_data[i] = 0;
	}
	pthread_rwlock_unlock(thread_lock);
//	pthread_rwlock_unlock(&lwm2_process_storage_lock);
}

//Merges all threads' complete execution metrics with process's complete execution data
void lwm2_storage_merge_threads_sum_data()
{
	int i;
	unsigned long long* thread_sum_data;
	
	for(i = 0; i < lwm2_omp_thread_count; i++)
	{
		thread_sum_data = lwm2_thread_storage_table[i].thread_data_sum;
		lwm2_storage_merge_thread_sum_data(thread_sum_data);
	}
}

//Merges thread's complete execution metrics with process's complete execution data.
void lwm2_storage_merge_thread_sum_data(unsigned long long* thread_sum_data)
{
	int i;
	
	if(thread_sum_data == NULL)
	{
		LWM2LOG(2, "[%d] The thread sum data is NULL!\n", lwm2_mpi_rank);
		return;
	}
	
	for(i = 0; i < LWM2_TOTAL_METRIC_COUNT; i++)
	{
		lwm2_metric_total[i] += thread_sum_data[i];
	}
}

//Create a new storage node for time-slice metrics
struct metric_node* new_metric_node()
{

	struct metric_node* new_node = (struct metric_node*) lwm2_memory_new_time_slice_storage_bytes(sizeof(struct metric_node));
	
	if(new_node == NULL)
	{
		fprintf(stderr, "Rank: %d, Couldn't allocate memory for node\n", lwm2_mpi_rank);
		return new_node;
	}
	new_node->next = NULL;
	return new_node;
}

//Init the time-slice memory data structure
void lwm2_metric_storage_init()
{
	current_node = list_head = new_metric_node();
}

//Increment storage for next time-slice
void lwm2_metric_storage_next()
{
	lwm2_time_slice_index++;
#ifndef LWM2_CLUSTRX_SOCKET

	struct metric_node* next_node = new_metric_node();
	current_node->next = next_node;
	current_node = next_node;

#endif
}

//Init storage data structure for a thread
void lwm2_storage_init_thread_storage(int index)
{
	//Allocate memory the size of a system page
//	void* buff = lwm2_memory_alloc(lwm2_system_page_size, LWM2_MEMORY_COLOR_PTHREADS);
	void* buff = lwm2_new_thread_storage();
			
	if(buff == NULL)
	{
		LWM2LOG(2, "ERROR! Can't allocate memory for thread");
		return;
	}
	else
	{
		memset(buff, 0, lwm2_system_page_size);
	}

	//Allocate different portions to different buffers
	unsigned long long* local_buff_sum = (unsigned long long*) buff;
	unsigned long long* local_buff_slice_1 = (unsigned long long*) &local_buff_sum[LWM2_TOTAL_METRIC_COUNT];
	unsigned long long* local_buff_slice_2 = (unsigned long long*) &local_buff_slice_1[LWM2_WRAPPER_COUNT];
	char* flags_buff = (char*) &local_buff_slice_2[LWM2_WRAPPER_COUNT];
	pthread_rwlock_t* thread_rw_lock = NULL;
#ifdef LWM2_MEASUREMENT_HWCOUNTER
#ifdef LWM2_PAPI_THREAD_ENABLED
	struct lwm2_hwc_storage* local_pthread_storage = (struct lwm2_hwc_storage*) &flags_buff[FLAG_SIGNAL_COUNT];
	thread_rw_lock = (pthread_rwlock_t*) (local_pthread_storage + sizeof(struct lwm2_hwc_storage));
#else
	thread_rw_lock	 = (pthread_rwlock_t*) &flags_buff[FLAG_SIGNAL_COUNT];
#endif
#else
	 thread_rw_lock	 = (pthread_rwlock_t*) &flags_buff[FLAG_SIGNAL_COUNT];
#endif
	
	pthread_rwlock_init(thread_rw_lock, NULL);

	lwm2_thread_storage_table[index].thread_data_sum = local_buff_sum;
	lwm2_thread_storage_table[index].thread_data_slice_1 = local_buff_slice_1;
	lwm2_thread_storage_table[index].thread_data_slice_2 = local_buff_slice_2;
	lwm2_thread_storage_table[index].thread_data_slice_index = 0;
	lwm2_thread_storage_table[index].my_flags = flags_buff;
	lwm2_thread_storage_table[index].did_read_hwc = 0;
#ifdef LWM2_MEASUREMENT_HWCOUNTER
#ifdef LWM2_PAPI_THREAD_ENABLED
	lwm2_thread_storage_table[index].hwc_storage = local_pthread_storage;
#endif
#endif
	lwm2_thread_storage_table[index].thread_lock = thread_rw_lock;

	pthread_setspecific(lwm2_pthread_data_key, &lwm2_thread_storage_table[index]);

}

void* lwm2_storage_get_thread_storage_pthread(int index)
{
	if(index < lwm2_omp_thread_count)
	{
#ifdef LWM2_MEASUREMENT_HWCOUNTER
#ifdef LWM2_PAPI_THREAD_ENABLED
		return (void*) lwm2_thread_storage_table[index].hwc_storage;
#endif
#else
		return (void*) NULL;
#endif
	}
	return (void*) NULL;
}

//Initialize storage for metrics
int lwm2_storage_init()
{
	int i;
	
	struct timeval tv;
	gettimeofday(&tv, NULL);
	
	pthread_rwlock_init(&lwm2_process_storage_lock, NULL);
	
	lwm2_metric_storage_init();
	
	for(i = 0; i < MAX_THREADS; i++)
	{
		lwm2_thread_storage_table[i].thread_data_sum = NULL;
		lwm2_thread_storage_table[i].thread_data_slice_1 = NULL;
		lwm2_thread_storage_table[i].thread_data_slice_2 = NULL;
		lwm2_thread_storage_table[i].thread_data_slice_index = 0;
		lwm2_thread_storage_table[i].my_flags = NULL;
		lwm2_thread_storage_table[i].thread_lock = NULL;
	}

	memset(current_node->metrics_data, 0, sizeof(unsigned long long) * LWM2_TOTAL_METRIC_COUNT);
	memset(lwm2_metric_total, 0 , sizeof(unsigned long long) * LWM2_TOTAL_METRIC_COUNT);
	
	//Initialize the Max/Min metrics with initial values
	//These are unsigned values, -1 results in the max possible value
	lwm2_mpi_comm_data[MIN_P2P_MSG_SIZE] = -1;
	lwm2_mpi_comm_data[MAX_P2P_MSG_SIZE] = 0;
	lwm2_mpi_comm_data[MIN_COLL_MSG_SENT] = -1;
	lwm2_mpi_comm_data[MIN_COLL_MSG_RECV] = -1;
	lwm2_mpi_comm_data[MAX_COLL_MSG_SENT] = 0;
	lwm2_mpi_comm_data[MAX_COLL_MSG_RECV] = 0;
	
	lwm2_system_page_size = sysconf(_SC_PAGE_SIZE);

	return LWM2_SUCCESS;
}

int lwm2_storage_reset()
{
	int i;

	for(i = 0; i < LWM2_WRAPPER_COUNT; i++)
	{
		current_node->metrics_data[i] = 0;
	}
	
	return LWM2_SUCCESS;
}

//Perform the necessary steps at the end of a time-slice
int lwm2_storage_update_time_slice()
{
	//Change the thread storage buffers
	lwm2_storage_change_thread_storage();
	//Merge the thread time-slice storage
	lwm2_storage_merge_threads_time_slice_data();
	//Send the data through sockets to ClustrX
	//This function doesn't do anything if ClustrX is not enabled at compile time
	lwm2_socket_send_data();
	//Increment the time-slice storage
	lwm2_metric_storage_next();
	return LWM2_SUCCESS;
}

//Perform the necessary steps at the end of a time-slice
//and reset the new memory storage for time-slice metrics
int lwm2_storage_update_time_slice_and_reset()
{
	
	lwm2_storage_update_time_slice();
	
	int i;
	
	for(i = 0; i < LWM2_TOTAL_METRIC_COUNT; i++)
	{
		current_node->metrics_data[i] = 0;
	}
	
	return LWM2_SUCCESS;
}

//Check if this metric is stored per time-slice or only for the whole execution
int lwm2_storage_is_metric_per_time_slice(unsigned char metric_id)
{
	//The starting metrics, till LWM2_THREAD_METRIC_COUNT are not stored per time-slice
	if(metric_id >= LWM2_THREAD_METRIC_COUNT && metric_id < LWM2_TOTAL_METRIC_COUNT)
	{
		return 1;
	}
	return 0;
}

//The current node in the linked-list is the current time-slice storage
unsigned long long* lwm2_storage_get_process_time_slice_storage()
{
	return current_node->metrics_data;
}

//Wrapper around the global storage
unsigned long long* lwm2_storage_get_process_global_storage()
{
	return lwm2_metric_total;
}

/*
 * Generic function, used by specific functions below. Handles per thread storage also.
 */
int lwm2_storage_set_metric(unsigned char metric_id, unsigned long long value)
{
	unsigned long long* time_slice_storage = NULL, *total_storage = NULL;
	total_storage = lwm2_storage_get_my_storage_sum();
	pthread_rwlock_t* thread_lock = NULL;
	
	//location to store metric in time-slice buffer is calculated by removing the number of thread specific storage
	int time_slice_metric_index = metric_id - LWM2_THREAD_METRIC_COUNT;

	if(metric_id >= LWM2_TOTAL_METRIC_COUNT)
	{
		LWM2LOG(3, "Received invalid key for storage: %d, ignoring\n", metric_id);
		return -1;
	}

	//Increment the metric for the complete execution
	total_storage[metric_id] += value;
	
	//if it is stored per time-slice
	if(lwm2_storage_is_metric_per_time_slice(metric_id))
	{
		//Acquire lock only when sure it is stored per time-slice
		time_slice_storage = lwm2_storage_get_my_storage_slice(thread_lock);
		
		//if there is no per-thread locking mechanism defined, use the process wide lock
		if(thread_lock != NULL)
		{
			pthread_rwlock_wrlock(thread_lock);
		}
		else
		{
			pthread_rwlock_wrlock(&lwm2_process_storage_lock);
		}

		//Set the value per time-slice
		time_slice_storage[time_slice_metric_index] = value;
		
		//Unlock the appropriate lock
		if(thread_lock != NULL)
		{
			pthread_rwlock_unlock(thread_lock);
		}
		else
		{
			pthread_rwlock_unlock(&lwm2_process_storage_lock);
		}
	}

	return LWM2_SUCCESS;
}

/*
 * Changes the value of the metric, only if the new value is greater
 */
int lwm2_storage_set_max_metric(unsigned char metric_id, unsigned long long value)
{

	unsigned long long* time_slice_storage = NULL, *total_storage = NULL;
	total_storage = lwm2_storage_get_my_storage_sum();
	pthread_rwlock_t* thread_lock = NULL;
	
	//Subtract the non-time-slice count to get the index per time-slice
	int time_slice_metric_index = metric_id - LWM2_THREAD_METRIC_COUNT;

	if(metric_id >= LWM2_TOTAL_METRIC_COUNT)
	{
		LWM2LOG(3, "Received invalid key for storage: %d, ignoring\n", metric_id);
		return -1;
	}

	//Change the value only if the new value is greater
	if(total_storage[metric_id] < value)
	{
		total_storage[metric_id] = value;
	}

	if(lwm2_storage_is_metric_per_time_slice(metric_id))
	{
		//Acquire lock only when sure it is stored per time-slice
		time_slice_storage = lwm2_storage_get_my_storage_slice(thread_lock);
		if(thread_lock != NULL)
		{
			pthread_rwlock_wrlock(thread_lock);
		}
		else
		{
			pthread_rwlock_wrlock(&lwm2_process_storage_lock);
		}

		if(time_slice_storage[time_slice_metric_index] < value)
		{
			time_slice_storage[time_slice_metric_index] = value;
		}

		if(thread_lock != NULL)
		{
			pthread_rwlock_unlock(thread_lock);
		}
		else
		{
			pthread_rwlock_unlock(&lwm2_process_storage_lock);
		}
	}

	return LWM2_SUCCESS;
}

/*
 * Add the new value to the current value of the metric
 */
int lwm2_storage_add_metric(unsigned char metric_id, unsigned long long value)
{

	unsigned long long* time_slice_storage = NULL, *total_storage = NULL;

	total_storage = lwm2_storage_get_my_storage_sum();
	pthread_rwlock_t* thread_lock = NULL;
	
	int time_slice_metric_index = metric_id - LWM2_THREAD_METRIC_COUNT;

	if(metric_id >= LWM2_TOTAL_METRIC_COUNT)
	{
		LWM2LOG(3, "Received invalid key for storage: %d, ignoring\n", metric_id);
		return -1;
	}

	total_storage[metric_id] += value;
	if(lwm2_storage_is_metric_per_time_slice(metric_id))
	{
		//Acquire lock only when sure it is stored per time-slice
		time_slice_storage = lwm2_storage_get_my_storage_slice(thread_lock);
		if(thread_lock != NULL)
		{
			pthread_rwlock_wrlock(thread_lock);
		}
		else
		{
			pthread_rwlock_wrlock(&lwm2_process_storage_lock);
		}

		time_slice_storage[time_slice_metric_index] += value;

		if(thread_lock != NULL)
		{
			pthread_rwlock_unlock(thread_lock);
		}
		else
		{
			pthread_rwlock_unlock(&lwm2_process_storage_lock);
		}
	}

	return LWM2_SUCCESS;
}

/*
 * Increment the value of the specified metric in the storage by one
 */
int lwm2_storage_increment_metric(unsigned char metric_id)
{
	unsigned long long* time_slice_storage = NULL, *total_storage = NULL;

	total_storage = lwm2_storage_get_my_storage_sum();
	pthread_rwlock_t* thread_lock = NULL;
	
	int time_slice_metric_index = metric_id - LWM2_THREAD_METRIC_COUNT;

	if(metric_id >= LWM2_TOTAL_METRIC_COUNT)
	{
		LWM2LOG(3, "Received invalid key for storage: %d, ignoring\n", metric_id);
		return -1;
	}

	total_storage[metric_id]++;

	if(lwm2_storage_is_metric_per_time_slice(metric_id))
	{
		//Acquire lock only when sure it is stored per time-slice
		time_slice_storage = lwm2_storage_get_my_storage_slice(thread_lock);
		if(thread_lock != NULL)
		{
			pthread_rwlock_wrlock(thread_lock);
		}
		else
		{
			pthread_rwlock_wrlock(&lwm2_process_storage_lock);
		}

		time_slice_storage[time_slice_metric_index]++;

		if(thread_lock != NULL)
		{
			pthread_rwlock_unlock(thread_lock);
		}
		else
		{
			pthread_rwlock_unlock(&lwm2_process_storage_lock);
		}
	}

	return LWM2_SUCCESS;
}

/*
 * Copy the current time-slice metrics to the passed buffer
 */
int lwm2_storage_copy_metrics(unsigned long long metrics_copy[LWM2_TOTAL_METRIC_COUNT])
{
	int i;
	
	pthread_rwlock_rdlock(&lwm2_process_storage_lock);
	
	for(i = 0; i < LWM2_WRAPPER_COUNT; i++)
	{
		metrics_copy[i] = current_node->metrics_data[i];
	}
	pthread_rwlock_unlock(&lwm2_process_storage_lock);
	
	return LWM2_SUCCESS;
}

int lwm2_storage_increment_signal_total()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_TOTAL);
}

int lwm2_storage_increment_signal_mpi()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_MPI);
}

int lwm2_storage_increment_signal_mpi_p2p()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_MPI_P2P);
}

int lwm2_storage_increment_signal_mpi_coll()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_MPI_COLL);
}

int lwm2_storage_increment_signal_fileio()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_FILE_IO);
}

int lwm2_storage_increment_signal_mpi_io()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_MPI_IO);
}

int lwm2_storage_increment_signal_cuda()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_CUDA);
}

int lwm2_storage_increment_signal_cuda_to_device()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_CUDA_TO_DEVICE);
}

int lwm2_storage_increment_signal_cuda_to_host()
{
	return lwm2_storage_increment_metric(LWM2_SIGNAL_CUDA_TO_HOST);
}

int lwm2_storage_set_max_threads(int thread_count)
{
	return lwm2_storage_set_metric(LWM2_METRIC_MAX_NUM_THREADS, thread_count);
}

int lwm2_storage_check_set_max_threads(int thread_count)
{
	return lwm2_storage_set_max_metric(LWM2_METRIC_MAX_NUM_THREADS, thread_count);
}

int lwm2_storage_add_p2p_sent_bytes(unsigned int bytes_sent)
{
	//Set the Max/Min before adding it to the storage
	if(lwm2_mpi_comm_data[MIN_P2P_MSG_SIZE] > bytes_sent)
	{
		lwm2_mpi_comm_data[MIN_P2P_MSG_SIZE] = bytes_sent;
	}
	if(lwm2_mpi_comm_data[MAX_P2P_MSG_SIZE] < bytes_sent)
	{
		lwm2_mpi_comm_data[MAX_P2P_MSG_SIZE] = bytes_sent;
	}
	return lwm2_storage_add_metric(LWM2_METRIC_P2P_SENT_BYTES, bytes_sent);
}

int lwm2_storage_increment_p2p_sent_msgs()
{
	return lwm2_storage_increment_metric(LWM2_METRIC_P2P_SENT_MSGS);
}

int lwm2_storage_add_p2p_recv_bytes(unsigned int bytes_recv)
{
	//Set the Max/Min before adding it to the storage
	if(lwm2_mpi_comm_data[MIN_P2P_MSG_SIZE] > bytes_recv)
	{
		lwm2_mpi_comm_data[MIN_P2P_MSG_SIZE] = bytes_recv;
	}
	if(lwm2_mpi_comm_data[MAX_P2P_MSG_SIZE] < bytes_recv)
	{
		lwm2_mpi_comm_data[MAX_P2P_MSG_SIZE] = bytes_recv;
	}
	return lwm2_storage_add_metric(LWM2_METRIC_P2P_RECV_BYTES, bytes_recv);
}

int lwm2_storage_increment_p2p_recv_msgs()
{
	return lwm2_storage_increment_metric(LWM2_METRIC_P2P_RECV_MSGS);
}

int lwm2_storage_add_coll_sent_bytes(unsigned int bytes_sent)
{
	//Set the Max/Min before adding it to the storage
	if(lwm2_mpi_comm_data[MIN_COLL_MSG_SENT] > bytes_sent)
	{
		lwm2_mpi_comm_data[MIN_COLL_MSG_SENT] = bytes_sent;
	}
	if(lwm2_mpi_comm_data[MAX_COLL_MSG_SENT] < bytes_sent)
	{
		lwm2_mpi_comm_data[MAX_COLL_MSG_SENT] = bytes_sent;
	}
	return lwm2_storage_add_metric(LWM2_METRIC_COLL_SENT_BYTES, bytes_sent);
}

int lwm2_storage_add_coll_recv_bytes(unsigned int bytes_recv)
{
	//Set the Max/Min before adding it to the storage
	if(lwm2_mpi_comm_data[MIN_COLL_MSG_RECV] > bytes_recv)
	{
		lwm2_mpi_comm_data[MIN_COLL_MSG_RECV] = bytes_recv;
	}
	if(lwm2_mpi_comm_data[MAX_COLL_MSG_RECV] < bytes_recv)
	{
		lwm2_mpi_comm_data[MAX_COLL_MSG_RECV] = bytes_recv;
	}
	return lwm2_storage_add_metric(LWM2_METRIC_COLL_RECV_BYTES, bytes_recv);
}

int lwm2_storage_increment_coll_calls()
{
	return lwm2_storage_increment_metric(LWM2_METRIC_COLL_CALL_COUNT);
}

int lwm2_storage_add_io_read_bytes(unsigned int bytes_read)
{
	return lwm2_storage_add_metric(LWM2_METRIC_IO_READ_BYTES, bytes_read);
}

int lwm2_storage_add_io_write_bytes(unsigned int bytes_write)
{
	return lwm2_storage_add_metric(LWM2_METRIC_IO_WRITE_BYTES, bytes_write);
}

int lwm2_storage_add_cuda_bytes_to_device(unsigned long bytes_to_device)
{
	return lwm2_storage_add_metric(LWM2_METRIC_CUDA_BYTES_TO_DEVICE, bytes_to_device);
}

int lwm2_storage_add_cuda_bytes_to_host(unsigned long bytes_to_host)
{
	return lwm2_storage_add_metric(LWM2_METRIC_CUDA_BYTES_TO_HOST, bytes_to_host);
}

int lwm2_storage_increment_msgs_to_device()
{
	return lwm2_storage_increment_metric(LWM2_METRIC_CUDA_TO_DEVICE_CALLS);
}

int lwm2_storage_increment_msgs_to_host()
{
	return lwm2_storage_increment_metric(LWM2_METRIC_CUDA_TO_HOST_CALLS);
}

int lwm2_storage_set_cuda_instructions(unsigned long long inst_count)
{
	return lwm2_storage_set_metric(LWM2_METRIC_CUDA_INSTRUCTIONS, inst_count);
}

int lwm2_storage_add_cuda_instructions(unsigned long long inst_count)
{
	return lwm2_storage_add_metric(LWM2_METRIC_CUDA_INSTRUCTIONS, inst_count);
}

int lwm2_storage_set_cuda_divergent_branches(unsigned long long div_branches)
{
	return lwm2_storage_set_metric(LWM2_METRIC_CUDA_DIVERGENT, div_branches);
}

int lwm2_storage_add_cuda_divergent_branches(unsigned long long div_branches)
{
	return lwm2_storage_add_metric(LWM2_METRIC_CUDA_DIVERGENT, div_branches);
}

#ifndef LWM2_HWC_EXTERN_DEF

//These specific functions are only for the default HWCs defined.
int lwm2_storage_set_cpu_instructions(unsigned long long inst_count)
{
	return lwm2_storage_set_metric(LWM2_METRIC_INSTRUCTIONS, inst_count);
}

int lwm2_storage_set_cpu_cycles(unsigned long long inst_cycles)
{
	return lwm2_storage_set_metric(LWM2_METRIC_CYCLES, inst_cycles);
}

int lwm2_storage_set_cpu_flops(unsigned long long flop)
{
	int ret1, ret2;
	unsigned long long flops = (unsigned long long) (flop / (long double) lwm2_time_slice_duration);
	ret1 = lwm2_storage_set_metric(LWM2_METRIC_FLOPS, flops);
	if(lwm2_hwc_registered & LWM2_HWC_FLOINS == 0)
	{
		ret2 = lwm2_storage_set_cpu_floins(flop);
	}
	else
	{
		ret2 = ret1;
	}
	
	return (ret1 & ret2);
}

int lwm2_storage_set_cpu_floins(unsigned long long floins)
{
	return lwm2_storage_set_metric(LWM2_METRIC_FLOINS, floins);
}

int lwm2_storage_set_l1_data_accesses(unsigned long long l1_data_accesses)
{
	return lwm2_storage_set_metric(LWM2_METRIC_L1_DATA_ACCESS, l1_data_accesses);
}

int lwm2_storage_set_l1_data_misses(unsigned long long l1_data_misses)
{
	return lwm2_storage_set_metric(LWM2_METRIC_L1_DATA_MISSES, l1_data_misses);
}

#else

//These are generic functions, called with external HWCs are defined
int lwm2_storage_set_custom_hwc(int hwc_id, unsigned long long hwc_value)
{
	//Add SIGNAL COUNT to get the proper id
	return lwm2_storage_set_metric(hwc_id + LWM2_SIGNAL_METRIC_COUNT, hwc_value);
}

int lwm2_storage_reset_custom_hwc(int hwc_id)
{
	//Add SIGNAL COUNT to get the proper id
	return lwm2_storage_set_metric(hwc_id + LWM2_SIGNAL_METRIC_COUNT, 0);
}

int lwm2_storage_set_custom_hwcs(unsigned long long* hwc_values)
{
	int i;
	int ret = 0;
	
	fprintf(stderr, "Storage: HWC count: %d\n", LWM2_HWC_METRIC_COUNT - LWM2_SIGNAL_METRIC_COUNT);
	
	for(i = 0; i < LWM2_HWC_METRIC_COUNT - LWM2_SIGNAL_METRIC_COUNT; i++)
	{
		//Add SIGNAL COUNT to get the proper id
		ret |= lwm2_storage_set_metric(i + LWM2_SIGNAL_METRIC_COUNT, hwc_values[i]);
	}
	
	return ret;
}

int lwm2_storage_reset_custom_hwcs()
{
	int i;
	int ret = 0;
	for(i = 0; i < LWM2_HWC_METRIC_COUNT - LWM2_SIGNAL_METRIC_COUNT; i++)
	{
		//Add SIGNAL COUNT to get the proper id
		ret |= lwm2_storage_set_metric(i + LWM2_SIGNAL_METRIC_COUNT, 0);
	}
	return ret;
}

#endif

int lwm2_storage_set_flag(int flag_type)
{
	char* my_flags = lwm2_storage_get_flag_storage();
	my_flags[flag_type] = 1;
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag(int flag_type)
{
	char* my_flags = lwm2_storage_get_flag_storage();
	if(my_flags == NULL)
	{
		return LWM2_ERROR;
	}
	my_flags[flag_type] = 0;
	return LWM2_SUCCESS;
}

int lwm2_storage_increment_flag(int flag_type)
{
	char* my_flags = lwm2_storage_get_flag_storage();
	my_flags[flag_type]++;
	return LWM2_SUCCESS;
}

int lwm2_storage_decrement_flag(int flag_type)
{
	char* my_flags = lwm2_storage_get_flag_storage();
	my_flags[flag_type]--;
	return LWM2_SUCCESS;
}

int lwm2_storage_get_flag(int flag_type)
{
	char* my_flags = lwm2_storage_get_flag_storage();
	return my_flags[flag_type];
}

int lwm2_storage_set_flag_sighandler()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_HANDLER_ACTIVE);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_mpi()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_MPI);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_mpi_p2p()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_MPI_P2P);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_mpi_coll()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_MPI_COLL);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_fileio()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_FILE_IO);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_mpi_io()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_MPI_IO);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_cuda()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_CUDA);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_cuda_to_device()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_CUDA_TO_DEVICE);
	return LWM2_SUCCESS;
}

int lwm2_storage_set_flag_cuda_to_host()
{
	lwm2_storage_set_flag(FLAG_SIGNAL_CUDA_TO_HOST);
	return LWM2_SUCCESS;
}

int lwm2_storage_increment_flag_cuda()
{
	lwm2_storage_increment_flag(FLAG_SIGNAL_CUDA);
	return LWM2_SUCCESS;
}

int lwm2_storage_get_flag_sighandler()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_HANDLER_ACTIVE);
}

int lwm2_storage_get_flag_mpi()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_MPI);
}

int lwm2_storage_get_flag_mpi_p2p()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_MPI_P2P);
}

int lwm2_storage_get_flag_mpi_coll()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_MPI_COLL);
}

int lwm2_storage_get_flag_fileio()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_FILE_IO);
}

int lwm2_storage_get_flag_mpi_io()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_MPI_IO);
}

int lwm2_storage_get_flag_cuda()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_CUDA);
}

int lwm2_storage_get_flag_cuda_to_device()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_CUDA_TO_DEVICE);
}

int lwm2_storage_get_flag_cuda_to_host()
{
	return lwm2_storage_get_flag(FLAG_SIGNAL_CUDA_TO_HOST);
}

int lwm2_storage_decrement_flag_cuda()
{
	lwm2_storage_decrement_flag(FLAG_SIGNAL_CUDA);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_sighandler()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_HANDLER_ACTIVE);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_mpi()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_MPI);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_mpi_p2p()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_MPI_P2P);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_mpi_coll()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_MPI_COLL);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_fileio()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_FILE_IO);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_mpi_io()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_MPI_IO);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_cuda()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_CUDA);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_cuda_to_device()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_CUDA_TO_DEVICE);
	return LWM2_SUCCESS;
}

int lwm2_storage_reset_flag_cuda_to_host()
{
	lwm2_storage_reset_flag(FLAG_SIGNAL_CUDA_TO_HOST);
	return LWM2_SUCCESS;
}

