/****************************************************************************
**  Copyright (c) 2011-2012                                                **
**  German Research School for Simulation Sciences GmbH					   **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef LWM2_STORAGE_H_
#define LWM2_STORAGE_H_

#include <pthread.h>

#include "lwm2-config.h"

#include "adapter/lwm2-hwc-adapter.h"
#include "lwm2-memory.h"

//List of signal based metrics
enum {
	LWM2_SIGNAL_TOTAL = 0,
	LWM2_SIGNAL_MPI,
	LWM2_SIGNAL_FILE_IO,
	LWM2_SIGNAL_MPI_IO,
	LWM2_SIGNAL_CUDA,
	LWM2_SIGNAL_CUDA_TO_DEVICE,
	LWM2_SIGNAL_CUDA_TO_HOST,
	LWM2_SIGNAL_MPI_P2P,
	LWM2_SIGNAL_MPI_COLL,
	LWM2_SIGNAL_METRIC_COUNT
};

#ifndef LWM2_HWC_EXTERN_DEF

//List of default HWC based metrics. This is used when an external list is not specified.
enum {
	LWM2_METRIC_CYCLES = LWM2_SIGNAL_METRIC_COUNT,
	LWM2_METRIC_FLOINS,
	LWM2_METRIC_FLOPS,
	LWM2_METRIC_INSTRUCTIONS,
	LWM2_METRIC_L1_DATA_ACCESS,
	LWM2_METRIC_L1_DATA_MISSES
};

#define LWM2_PREDEFINED_HWC_COUNT 6

//This is used to keep the metric list a continuous ascending numerical list
#define LWM2_HWC_METRIC_COUNT (LWM2_SIGNAL_METRIC_COUNT + LWM2_PREDEFINED_HWC_COUNT)
#else 
//Include the external list of hardware counters
#include "modules/hwc/lwm2-hwc-num.h"
//Set the value to make the enum a continuous ascending numerial list
#define LWM2_HWC_METRIC_COUNT (LWM2_SIGNAL_METRIC_COUNT + LWM2_CUSTOM_STORAGE_HWC_COUNT)
#endif


enum {
	LWM2_METRIC_P2P_SENT_BYTES = LWM2_HWC_METRIC_COUNT,
	LWM2_METRIC_P2P_RECV_BYTES,
	LWM2_METRIC_COLL_SENT_BYTES,
	LWM2_METRIC_COLL_RECV_BYTES,
	LWM2_METRIC_P2P_SENT_MSGS,
	LWM2_METRIC_P2P_RECV_MSGS,
	LWM2_METRIC_COLL_CALL_COUNT,
	LWM2_METRIC_IO_READ_BYTES,
	LWM2_METRIC_IO_WRITE_BYTES,
	LWM2_METRIC_CUDA_BYTES_TO_DEVICE,
	LWM2_METRIC_CUDA_BYTES_TO_HOST,
	LWM2_METRIC_CUDA_TO_DEVICE_CALLS,
	LWM2_METRIC_CUDA_TO_HOST_CALLS,
	LWM2_METRIC_CUDA_INSTRUCTIONS,
	LWM2_METRIC_CUDA_DIVERGENT,
	LWM2_METRIC_MAX_NUM_THREADS,
	LWM2_TOTAL_METRIC_COUNT
};

//Specify the metrics that are to be stored per time-slice
enum {
#ifdef LWM2_PAPI_THREAD_ENABLED
	LWM2_THREAD_METRIC_COUNT = LWM2_HWC_METRIC_COUNT
#else
	LWM2_THREAD_METRIC_COUNT = LWM2_SIGNAL_METRIC_COUNT
#endif
};

//List of flags set by wrappers and checked in signal hit
enum {
	FLAG_SIGNAL_MPI = 0,
	FLAG_SIGNAL_FILE_IO,
	FLAG_SIGNAL_MPI_IO,
	FLAG_SIGNAL_CUDA,
	FLAG_SIGNAL_CUDA_TO_DEVICE,
	FLAG_SIGNAL_CUDA_TO_HOST,
	FLAG_SIGNAL_MPI_P2P,
	FLAG_SIGNAL_MPI_COLL,
	FLAG_SIGNAL_HANDLER_ACTIVE,
	FLAG_SIGNAL_COUNT
};

//Metrics that are stored per time-slice, and are not thread dependent
#define LWM2_WRAPPER_COUNT (LWM2_TOTAL_METRIC_COUNT - LWM2_THREAD_METRIC_COUNT)

//Make metric count visible to other files
extern const int lwm2_metric_count;

//Make metric names visible to other files
extern char* lwm2_signal_metric_names[];
extern char* lwm2_hwc_metric_names[];
extern char* lwm2_tsliced_metric_names[];

//Make the table with host names visible to other files
extern char* lwm2_host_table;

//Some values needed when displaying summary
enum {
	MIN_P2P_MSG_SIZE = 0,
	MAX_P2P_MSG_SIZE,
	MIN_COLL_MSG_SENT,
	MIN_COLL_MSG_RECV,
	MAX_COLL_MSG_SENT,
	MAX_COLL_MSG_RECV,
	COMM_DATA_TOTAL
};

//Make metrics data visible to other modules
extern unsigned long long lwm2_mpi_comm_data[COMM_DATA_TOTAL];
extern unsigned long long lwm2_metric_total[LWM2_TOTAL_METRIC_COUNT];

//Storage for individual threads
struct lwm2_thread_storage {
	unsigned long long* thread_data_sum;
	unsigned long long* thread_data_slice_1;
	unsigned long long* thread_data_slice_2;
	int thread_data_slice_index;
	char* my_flags;
	int did_read_hwc;
//if hardware counters should be read
#ifdef LWM2_MEASUREMENT_HWCOUNTER
//if hardware counters are to be managed per thread	
#ifdef LWM2_PAPI_THREAD_ENABLED
	struct lwm2_hwc_storage* hwc_storage;
#endif
	
#endif
	pthread_rwlock_t* thread_lock;
};

//storage for thread data
extern struct lwm2_thread_storage lwm2_thread_storage_table[MAX_THREADS];
//lock for thread storage
extern pthread_rwlock_t lwm2_thread_storage_table_lock;

//time-slice count
extern long lwm2_time_slice_index;

//flags used in signal hit
char lwm2_all_flags[FLAG_SIGNAL_COUNT];

//Data storage for one time-slice
struct metric_node {
	unsigned long long metrics_data[LWM2_WRAPPER_COUNT];
	struct metric_node* next;
};

//the head node of time-slice's linked list
extern struct metric_node* list_head;

struct metric_node* new_metric_node();

void lwm2_metric_storage_init();

void lwm2_storage_init_thread_storage(int index);

void* lwm2_storage_get_thread_storage_pthread(int index);

unsigned long long* lwm2_storage_get_my_storage_sum();

unsigned long long* lwm2_storage_get_my_storage_slice(pthread_rwlock_t* thread_lock);

void* lwm2_storage_get_my_hwc_storage();

void lwm2_storage_reset_my_hwc_storage();

int lwm2_storage_am_i_a_registered_thread();

void lwm2_metric_storage_next();

void lwm2_storage_change_thread_storage();
void lwm2_storage_merge_threads_time_slice_data();
void lwm2_storage_merge_thread_time_slice_data(unsigned long long* thread_data, unsigned long long* process_data, pthread_rwlock_t* thread_lock);
void lwm2_storage_merge_threads_sum_data();
void lwm2_storage_merge_thread_sum_data(unsigned long long* thread_sum_data);

int lwm2_storage_init();

int lwm2_storage_update_time_slice();

int lwm2_storage_reset();

int lwm2_storage_update_time_slice_and_reset();

int lwm2_storage_switch_thread_buffers();

int lwm2_storage_is_metric_per_time_slice(unsigned char metric_id);

unsigned long long* lwm2_storage_get_process_time_slice_storage();

unsigned long long* lwm2_storage_get_process_global_storage();

int lwm2_storage_set_thread_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_set_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_set_max_thread_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_set_max_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_add_thread_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_add_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_increment_thread_metric(unsigned char metric_id);

int lwm2_storage_increment_metric(unsigned char metric_id);

int lwm2_storage_copy_metrics(unsigned long long metrics_copy[LWM2_TOTAL_METRIC_COUNT]);

int lwm2_storage_add_total_metric(unsigned char metric_id, unsigned long long value);

int lwm2_storage_increment_total_metric(unsigned char metric_id);

//int lwm2_storage_merge_thread_metrics();

int lwm2_storage_increment_signal_total();

//int lwm2_storage_increment_signal_serial();

int lwm2_storage_increment_signal_mpi();

int lwm2_storage_increment_signal_mpi_p2p();

int lwm2_storage_increment_signal_mpi_coll();

int lwm2_storage_increment_signal_fileio();

int lwm2_storage_increment_signal_mpi_io();

//int lwm2_storage_increment_signal_non_omp_thread();

//int lwm2_storage_increment_signal_grey_area_thread();

int lwm2_storage_increment_signal_cuda();

int lwm2_storage_increment_signal_cuda_to_device();

int lwm2_storage_increment_signal_cuda_to_host();

int lwm2_storage_set_max_threads(int thread_count);

int lwm2_storage_check_set_max_threads(int thread_count);

int lwm2_storage_add_p2p_sent_bytes(unsigned int bytes_sent);

int lwm2_storage_increment_p2p_sent_msgs();

int lwm2_storage_add_p2p_recv_bytes(unsigned int bytes_recv);

int lwm2_storage_increment_p2p_recv_msgs();

int lwm2_storage_add_coll_sent_bytes(unsigned int bytes_sent);

int lwm2_storage_add_coll_recv_bytes(unsigned int bytes_recv);

int lwm2_storage_increment_coll_calls();

int lwm2_storage_add_io_read_bytes(unsigned int bytes_read);

int lwm2_storage_add_io_write_bytes(unsigned int bytes_write);

int lwm2_storage_add_cuda_bytes_to_device(unsigned long bytes_to_device);

int lwm2_storage_add_cuda_bytes_to_host(unsigned long bytes_to_host);

int lwm2_storage_increment_msgs_to_device();

int lwm2_storage_increment_msgs_to_host();

int lwm2_storage_set_cuda_instructions(unsigned long long inst_count);

int lwm2_storage_add_cuda_instructions(unsigned long long inst_count);

int lwm2_storage_set_cuda_divergent_branches(unsigned long long div_branches);

int lwm2_storage_add_cuda_divergent_branches(unsigned long long div_branches);

#ifndef LWM2_HWC_EXTERN_DEF

int lwm2_storage_set_cpu_instructions(unsigned long long inst_count);

int lwm2_storage_set_cpu_cycles(unsigned long long inst_cycles);

int lwm2_storage_set_cpu_flops(unsigned long long flop);

int lwm2_storage_set_cpu_floins(unsigned long long floins);

int lwm2_storage_set_l1_data_accesses(unsigned long long l1_data_accesses);

int lwm2_storage_set_l1_data_misses(unsigned long long l1_data_misses);

#else

int lwm2_storage_set_custom_hwc(int hwc_id, unsigned long long hwc_value);

int lwm2_storage_reset_custom_hwc(int hwc_id);

int lwm2_storage_set_custom_hwcs(unsigned long long* hwc_values);

int lwm2_storage_reset_custom_hwcs();

#endif

char* lwm2_storage_get_flag_storage();

int lwm2_storage_set_flag(int flag_type);

int lwm2_storage_reset_flag(int flag_type);

int lwm2_storage_increment_flag(int flag_type);

int lwm2_storage_decrement_flag(int flag_type);

int lwm2_storage_get_flag(int flag_type);

int lwm2_storage_set_flag_sighandler();

int lwm2_storage_set_flag_mpi();

int lwm2_storage_set_flag_mpi_p2p();

int lwm2_storage_set_flag_mpi_coll();

int lwm2_storage_set_flag_fileio();

int lwm2_storage_set_flag_mpi_io();

int lwm2_storage_set_flag_cuda();

int lwm2_storage_set_flag_cuda_to_device();

int lwm2_storage_set_flag_cuda_to_host();

int lwm2_storage_increment_flag_cuda();

int lwm2_storage_get_flag_sighandler();

int lwm2_storage_get_flag_mpi();

int lwm2_storage_get_flag_mpi_p2p();

int lwm2_storage_get_flag_mpi_coll();

int lwm2_storage_get_flag_fileio();

int lwm2_storage_get_flag_mpi_io();

int lwm2_storage_get_flag_cuda();

int lwm2_storage_get_flag_cuda_to_device();

int lwm2_storage_get_flag_cuda_to_host();

int lwm2_storage_decrement_flag_cuda();

int lwm2_storage_reset_flag_sighandler();

int lwm2_storage_reset_flag_mpi();

int lwm2_storage_reset_flag_mpi_p2p();

int lwm2_storage_reset_flag_mpi_coll();

int lwm2_storage_reset_flag_fileio();

int lwm2_storage_reset_flag_mpi_io();

int lwm2_storage_reset_flag_cuda();

int lwm2_storage_reset_flag_cuda_to_device();

int lwm2_storage_reset_flag_cuda_to_host();

#endif
